# Read and write data from disk.
#
# A dataset is a single folder. To get data from the folder, you need to create
# a file called "getters.py" containing functions that return data and
# metadata.
#
# Authors: vivekv2@gmail.com

from functools import lru_cache, partial
import importlib.util
import json
from pathlib import Path
from typing import Optional, Callable
from types import ModuleType
import h5py
import numpy as np

from ...utils import io as zeir


def _module_name_from_path(p: Path) -> str:
    """Return a globally unique module name associated with a given
    directory."""
    return str(p / "getters")


@lru_cache()
def _getter_module_from_path(dataset: Path) -> Optional[ModuleType]:
    module_name = _module_name_from_path(dataset)
    getter_file = dataset / "getters.py"

    if not getter_file.is_file():
        return None

    spec = importlib.util.spec_from_file_location(module_name, getter_file)

    getters = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(getters)  # type: ignore

    return getters


def _make_getter_with_default(name: str, default: Callable) -> Callable:
    """This hideous function creates a getter with the given name. If that
    function doesn't exist in a "getter.py" file within the dataset, then
    a specified default function is called. The default is bound with
    the path."""

    def _getter(p: Path, *getter_args):

        getters = _getter_module_from_path(p)

        if getters is None or not hasattr(getters, name):
            _fn = partial(default, p)

        else:
            _fn = partial(getters.__getattribute__(name), p)

        return _fn(*getter_args)

    _getter.__name__ = name

    return _getter


def _get_metadata_default(dataset_path: Path):
    # json_filename = dataset_path / "metadata.json"
    # with open(json_filename) as json_file:
    #     metadata = json.load(json_file)
    # return metadata
    return zeir.get_metadata(dataset_path)


get_metadata = _make_getter_with_default("get_metadata", _get_metadata_default)


def _get_slice_default(dataset_path: Path, t: int) -> np.ndarray:
    # """Return a slice from the "data" field of an HDF5 file.
    # Time (t) is assumed to be the last (longest-stride) dimension. This should
    # generally be cached to avoid opening and closing files too many times:
    #
    #     my_get_slice = lru_cache(get_h5_slice, maxsize=1000)
    #
    # """
    # h5_filename = dataset_path / "data.h5"
    #
    # f = h5py.File(h5_filename, 'r')
    #
    # return f["data"][t]
    return zeir.get_slice(dataset_path, t)


def save_metadata_default(x: dict, dataset_path: Path) -> None:
    """Save the given dictionary as metadata.json in the given path."""

    json_filename = dataset_path / "metadata.json"
    with open(json_filename, 'w') as json_file:
        json.dump(x, json_file)


def save_dataset_default(x: np.ndarray, dataset_path: Path):
    """Save the given array in the given folder."""

    h5_filename = dataset_path / "data.h5"
    if h5_filename.is_file():
        h5_filename.unlink()
    f = h5py.File(h5_filename, "w")

    chunk_size = (1, *x.shape[1:])
    dset = f.create_dataset("data", data=x, chunks=chunk_size,
                            compression="gzip", compression_opts=7)

    f.close()


get_slice = _make_getter_with_default("get_slice", _get_slice_default)


def get_slice_3D(dataset: Path, t: int, method=np.max) -> np.ndarray:
    """Return a 3D slice by projeting extra dimensions (usually color/channel).
    By default, this will provide a maximum intensity projection, but other
    functions with the same type signature as numpy.max will work as well."""

    A = get_slice(dataset, t)

    if np.ndim(A) == 3:
        return A

    return method(A, axis=0)


def get_channel_specific_slice_3D(dataset: Path, t: int, method=np.max, channel="*") -> np.ndarray:
    """Return a 3D slice for a specific channel.
    By default, this will provide a maximum intensity projection, but other
    functions with the same type signature as numpy.max will work as well."""

    A = get_slice(dataset, t)

    if np.ndim(A) == 3:
        return A
    else:
        if channel == "*":
            return method(A, axis=0)
        else:
            return A[int(channel)]


def _get_times_default(dataset_path: Path) -> np.ndarray:
    """Return the timestamp of a given data frame."""
    h5_filename = dataset_path / "data.h5"

    f = h5py.File(h5_filename, 'r')

    return f["times"][:]


get_times = _make_getter_with_default("get_times", _get_times_default)



# These are all the procedures that appear in the annotator.

from .delete_annotations import delete_annotations
from .insert_local_max import insert_local_max
from .insert_annotation import insert_annotation
from .worldline_utilities import randomly_color_worldlines, renumber_worldlines
from .navigation import jump_to_frame
from .create_track import create_track
from .change_provenance import change_provenance
from .update_frame import update_frame
from .overwrite_zeir_checkpoint import overwrite_zeir_checkpoint

__all__ = [
    "delete_annotations", "insert_local_max", "insert_annotation",
    "randomly_color_worldlines", "renumber_worldlines", "jump_to_frame",
    "create_track", "change_provenance", "update_frame", "overwrite_zeir_checkpoint"
]



import torch
import torch.nn as nn
from torch.nn.functional import grid_sample
from ..utils.utils import get_gaussian_kernel


class ZephIR(nn.Module):
    def __init__(self,
                 allow_rotation,
                 dimmer_ratio,
                 fovea_sigma,
                 grid_shape,
                 grid_spacing,
                 n_chunks,
                 n_frame,
                 shape_n,
                 ftr_ratio=0.6,
                 ret_stride=2,):

        super(ZephIR, self).__init__()

        self.n_chunks = max(n_chunks, 1)
        self.n_frame = n_frame
        self.shape_n = shape_n

        # calculating descriptor sampling stride
        # foveated region at the center has full resolution (stride=1)
        # retinal region at the edges has reduced resolution (stride=ret_stride)
        grid_fovea = []
        dimmer = []
        for dim, reach in zip(grid_shape, grid_spacing):
            fovea_idx = int((1 - ftr_ratio) * dim / 2)
            retina = torch.zeros(dim)
            dimmer_ = torch.ones(dim)
            if fovea_idx > 0:
                for i in range(fovea_idx + 1):
                    retina[fovea_idx - i] += -(ret_stride - 1) * (reach * 2 / dim) * i
                    retina[-(fovea_idx - i + 1)] += (ret_stride - 1) * (reach * 2 / dim) * i
                    dimmer_[fovea_idx - i] *= dimmer_ratio
                    dimmer_[-(fovea_idx - i + 1)] *= dimmer_ratio
            grid_fovea.append(retina)
            dimmer.append(dimmer_)
        x_dimmer, y_dimmer, z_dimmer = torch.meshgrid(dimmer, indexing='ij')

        # calculating Gaussian mask to modulate intensity
        # center has full intensity, with Gaussian roll-off towards the edges
        # the multiplier never drops below dimmer_ratio
        if fovea_sigma[-1] < 0.:
            fovea_kernel = torch.ones(grid_shape)
        else:
            fovea_kernel = get_gaussian_kernel(grid_shape, fovea_sigma, channels=0)
            fovea_kernel = fovea_kernel / torch.max(fovea_kernel)

        # consolidates foveation effects
        self.dimmer = nn.Parameter(
            torch.clamp(
                torch.clamp(x_dimmer*y_dimmer*z_dimmer, dimmer_ratio, 1.0) * fovea_kernel,
                dimmer_ratio, 1.0
            ), requires_grad=False
        )

        # pre-compiling an identity sampling grid
        z_grid, y_grid, x_grid = torch.meshgrid(
            torch.linspace(-1, 1, grid_shape[0]) * grid_spacing[0] + grid_fovea[0],
            torch.linspace(-1, 1, grid_shape[1]) * grid_spacing[1] + grid_fovea[1],
            torch.linspace(-1, 1, grid_shape[2]) * grid_spacing[2] + grid_fovea[2],
            indexing='ij'
        )
        grid = torch.stack((x_grid, y_grid, z_grid), dim=-1)
        self.grid = nn.Parameter(
            grid.expand(self.n_frame, self.shape_n, -1, -1, -1, -1),
            requires_grad=False
        )

        # transformation parameter
        # direct correspondence to keypoint coordinates
        self.rho = nn.Parameter(
            torch.zeros((self.n_frame, self.shape_n, 3)),
            requires_grad=True
        )

        # transformation parameter
        # controls rotation of descriptors about center in the xy-plane
        self.theta = nn.Parameter(
            torch.zeros((self.n_frame, self.shape_n, 1)),
            requires_grad=allow_rotation
        )

    def forward(self, input_tensor):

        # applying rotation
        s_theta = torch.sin(self.theta).view(self.n_frame, self.shape_n, 1, 1, 1)
        cs_theta = torch.cos(self.theta).view(self.n_frame, self.shape_n, 1, 1, 1)
        theta = torch.stack((
            self.grid[..., 0] * cs_theta - self.grid[..., 1] * s_theta,
            self.grid[..., 0] * s_theta + self.grid[..., 1] * cs_theta,
            self.grid[..., 2]
        ), dim=-1)

        output_list = []
        for i in range(input_tensor.shape[0]):

            # forward pass is divided into n_chunks to reduce memory consumption
            for _theta, _rho in \
                zip(torch.chunk(theta[i], self.n_chunks, dim=0),
                    torch.chunk(self.rho[i], self.n_chunks, dim=0)):

                output_list.append(
                    grid_sample(
                        input_tensor[i].expand(_rho.shape[0], -1, -1, -1, -1),
                        _theta + _rho.view(-1, 1, 1, 1, 3),         # applying translation
                        align_corners=False, padding_mode='border'
                    )
                )

        output_tensor = torch.cat(output_list, dim=0).view(
            self.n_frame, self.shape_n,
            input_tensor.shape[2], *self.dimmer.shape
        )
        return output_tensor * self.dimmer.expand_as(output_tensor)



import shutil

from ..utils.utils import *


def save_annotations(
    container,
    results,
    save_mode,):
    """Save results to file.

    Handles tracking results and compiles to an annotations data frame to save
    to file according to save_mode: 'o' will overwrite existing annotations.h5,
    'w' will write to coordinates.h5. Existing annotations will overwrite
    ZephIR results if include_all is True.

    :param container: variable container, needs to contain: dataset, exclude_self,
    exclusive_prov, include_all, p_list, provenance, shape_t, shape_n, worldline_id
    :param results: tracking results
    :param save_mode: mode for writing to file
    """

    # pull variables from container
    dataset = container.get('dataset')
    exclude_self = container.get('exclude_self')
    exclusive_prov = container.get('exclusive_prov')
    include_all = container.get('include_all')
    p_list = container.get('p_list')
    provenance = container.get('provenance')
    shape_t = container.get('shape_t')
    shape_n = container.get('shape_n')
    worldline_id = container.get('worldline_id')

    annotation = get_annotation_df(dataset)
    
    # saving result to .h5
    print('\nCompiling and saving results to file...')
    p_list = np.array(p_list)
    p_list[np.where(p_list==-1)] = np.where(p_list==-1)
    xyz_pd = np.concatenate(
        (np.repeat(np.arange(shape_t), shape_n)[:, np.newaxis],
         results.reshape((-1, 3)) / 2.0 + 0.5,
         np.tile(worldline_id, shape_t)[:, np.newaxis],
         np.repeat(p_list, shape_n)[:, np.newaxis],
         provenance.reshape((-1, 1))),
        axis=-1
    )

    if include_all:
        for t in np.unique(annotation['t_idx']):
            u, annot, prov = get_annotation(annotation, t, exclusive_prov, exclude_self)
            w_idx = np.where(
                np.logical_and(
                    xyz_pd[t * shape_n:(t+1) * shape_n, 0].astype(np.float32) == t,
                    np.isin(xyz_pd[t * shape_n:(t+1) * shape_n, 4].astype(np.float32), u))
            )[0] + t * shape_n
            if len(w_idx) > 0:
                u_idx = np.where(np.isin(u, xyz_pd[w_idx, 4].astype(np.float32)))[0]
                xyz_pd[w_idx, :] = np.concatenate(
                    (np.ones((len(u_idx), 1)) * t,
                     annot[u_idx, ...] / 2 + 0.5,
                     u[u_idx, np.newaxis],
                     np.ones((len(u_idx), 1)) * p_list[t],
                     prov[u_idx, np.newaxis]),
                    axis=-1
                )
            if exclusive_prov is not None or exclude_self is True:
                u, annot, prov = get_annotation(annotation, t, None, False)
                w_idx = np.where(
                    np.logical_and(
                        xyz_pd[t * shape_n:(t + 1) * shape_n, 0].astype(np.float32) == t,
                        np.isin(xyz_pd[t * shape_n:(t + 1) * shape_n, 4].astype(np.float32), u))
                )[0] + t * shape_n
            if len(u) > len(w_idx):
                u_idx = np.where(np.isin(u, xyz_pd[w_idx, 4].astype(np.float32), invert=True))[0]
                xyz_pd = np.append(
                    xyz_pd,
                    np.concatenate(
                        (np.ones((len(u_idx), 1)) * t,
                         annot[u_idx, ...] / 2 + 0.5,
                         u[u_idx, np.newaxis],
                         np.ones((len(u_idx), 1)) * p_list[t],
                         prov[u_idx, np.newaxis]),
                        axis=-1
                    ),
                    axis=0
                )

    columns = {
        't_idx': np.uint32,
        'x': np.float32,
        'y': np.float32,
        'z': np.float32,
        'worldline_id': np.uint32,
        'parent_id': np.uint32,
        'provenance': np.dtype("S4"),
    }
    if save_mode == 'o':
        if not (dataset / 'backup').is_dir():
            Path.mkdir(dataset / 'backup')
        now = datetime.datetime.now()
        now_ = now.strftime("%m_%d_%Y_%H_%M")
        shutil.copy(dataset / 'annotations.h5',
                    dataset / 'backup' / f'annotations_{now_}.h5')
        f = h5py.File(dataset / 'annotations.h5', mode='w')
    else:
        f = h5py.File(dataset / 'coordinates.h5', mode=save_mode)

    data = np.array(list(range(1, xyz_pd.shape[0] + 1)), dtype=np.uint32)
    f.create_dataset('id', shape=(xyz_pd.shape[0], ), dtype=np.uint32, data=data)

    for i, c in enumerate(columns.keys()):
        if c == 'provenance':
            data = np.array(xyz_pd[:, i], dtype=columns[c])
        else:
            data = np.array(xyz_pd[:, i].astype(np.float32), dtype=columns[c])
        f.create_dataset(c, shape=(xyz_pd.shape[0], ), dtype=columns[c], data=data)

    f.close()

    return



from dataclasses import asdict
from logging import Logger
from pathlib import Path
from typing import List
from tqdm import tqdm

from ..data import AnnotationTable, WorldlineTable, Annotation
from ...methods.track_frame import track_frame

from ._utilities import default_args


@default_args("_now, True, False, False")
def update_frame(
    dataset: Path,
    annotations: AnnotationTable,
    worldlines: WorldlineTable,
    window_state,
    arg: str,
    logger: Logger
) -> List[dict]:
    """
    arg: time, restrict_update, recompile_model, update_children
        time: Time to run ZephIR on.
        restrict_update: Update only around partial manual annotations.
        recompile_model: Rerun argument interpretation and resource compilation steps in ZephIR.
        update_children: Update results for all children frames.
    """

    arg_list = arg.replace(" ", "").split(",")

    if arg_list[0] == '_now':
        t_idx = window_state['t_idx']
    else:
        t_idx = int(arg_list[0])
    restrict_update = arg_list[1] in ['True', 'Y', 'y']
    recompile_model = arg_list[2] in ['True', 'Y', 'y']
    update_children = arg_list[3] in ['True', 'Y', 'y']

    ret = track_frame(
        dataset=Path(dataset),
        annotation=annotations.df,
        t_idx=t_idx,
        restrict_update=restrict_update,
        recompile_model=recompile_model,
        update_children=update_children
    )
    if ret is None:
        return [{"type": "annotations/get_annotations"}]
    else:
        t_list, results, worldline_id, provenance = ret

    for t in tqdm(range(len(t_list)), desc='Updating annotations', unit='frames'):
        annot_t = annotations.get_t(t_list[t])
        ids = list(annot_t.df['id'])
        wlid = []
        for i in ids:
            annot = asdict(annot_t.get(i))
            wlid.append(annot['worldline_id'])
        for i, w in enumerate(worldline_id):
            if w not in wlid:
                A = Annotation(
                    t_idx=window_state["t_idx"],
                    x=(results[t, i, 0] + 1) / 2,
                    y=(results[t, i, 1] + 1) / 2,
                    z=(results[t, i, 2] + 1) / 2,
                    worldline_id=w,
                    parent_id=0,
                    provenance=b'ZEIR'
                )
                annotations.insert(A)
            else:
                idx = wlid.index(w)
                annotations.update(
                    ids[idx],
                    {'x': (results[t, i, 0] + 1) / 2,
                     'y': (results[t, i, 1] + 1) / 2,
                     'z': (results[t, i, 2] + 1) / 2,
                     'provenance': provenance[t, i]}
                )

    return [{"type": "annotations/get_annotations"}]



from tqdm.notebook import tqdm

from ..models.zephir import ZephIR
from ..models.losses import *
from ..utils.utils import *


def plot_loss_maps(
    keypoint_to_visualize,
    frame_to_visualize,
    reference,
    map_resolution,
    map_size,
    trajectory,
    container,
    dimmer_ratio,
    grid_shape,
    fovea_sigma,
    n_chunks,
    zephod,
    lambda_d,
    lambda_n,
    lambda_n_mode,):

    # pull variables from container
    dataset = container.get('dataset')
    annot = container.get('annot')
    channel = container.get('channel')
    covar = container.get('covar')
    dev = container.get('dev')
    gamma = container.get('gamma')
    img_shape = np.array(container.get('img_shape'))
    neighbors = container.get('neighbors')
    shape_n = container.get('shape_n')
    t_annot = container.get('t_annot')

    z_stacks = min(img_shape[0], grid_shape[0])
    grid_shape = (z_stacks - (z_stacks+1) % 2, grid_shape[1], grid_shape[2])
    grid_spacing = tuple(np.array(grid_shape) / img_shape)
    model_kwargs = {
        'allow_rotation': False,
        'dimmer_ratio': dimmer_ratio,
        'fovea_sigma': fovea_sigma,
        'grid_shape': grid_shape,
        'grid_spacing': grid_spacing,
        'n_chunks': n_chunks,
        'n_frame': 2,
        'shape_n': shape_n,
        'ftr_ratio': 0.6,
        'ret_stride': 2,
    }
    zephir = ZephIR(**model_kwargs).to(dev)

    t = frame_to_visualize
    root = np.where(t_annot == reference)[0][0]
    target = np.where(t_annot == frame_to_visualize)[0][0]

    t_patch = [t, reference]

    xv, yv = np.meshgrid(
        np.linspace(-map_size//2, map_size//2, map_resolution),
        np.linspace(-map_size//2, map_size//2, map_resolution)
    )

    losses_at_trajectory = np.zeros((map_resolution, map_resolution, 4))
    losses_at_annotation = losses_at_trajectory.copy()

    with torch.no_grad():

        # compiling Zephod results for feature detection loss, L_D
        input_tensor_d = []
        if lambda_d > 0 and zephod is not None:
            for _t in t_patch:
                data = get_data(dataset, _t, g=gamma, c=channel)
                _pred = zephod(data)
                _pred = (torch.max(_pred) - _pred) / torch.max(_pred)
                input_tensor_d.append(torch.where(_pred < 0.5, 0., 0.5))
            input_tensor_d = torch.stack(input_tensor_d, dim=0)
            input_tensor_d = blur3d(input_tensor_d, (1, 5, 5), (1, 2, 2), dev=dev)
            input_tensor_d.requires_grad = False

        # compiling target descriptors from reference frame for image registration, L_R
        data = get_data(dataset, t_annot[root], g=gamma, c=channel)
        vol = to_tensor(data, n_dim=5, dev=dev)
        zephir.theta.zero_()
        zephir.rho.zero_()
        zephir.rho[:len(t_patch)].add_(
            to_tensor(annot[root], dev=dev).expand(len(t_patch), -1, -1)
        )
        target_tensor = zephir(vol.expand(len(t_patch), -1, -1, -1, -1, -1))

        # loading input volumes
        input_tensor = []
        for _t in t_patch:
            data = get_data(dataset, _t, g=gamma, c=channel)
            vol = to_tensor(data, n_dim=5, grad=False, dev=dev)
            input_tensor.append(vol)
        input_tensor = torch.stack(input_tensor, dim=0)

        # compiling spring connections and stiffnesses for spatial regularization, L_N
        ind = neighbors[root][:, :, 0]
        d_ref = to_tensor(neighbors[root][:, :, 1:], dev=dev)
        if covar is not None:
            nn_covar = np.empty_like(ind)
            for k in range(ind.shape[1]):
                nn_covar[:, k] = covar[range(ind.shape[0]), ind[:, k].astype(int)] / 20
            k_ij = to_tensor(
                nn_covar /
                np.clip(np.max(nn_covar, axis=-1)[:, None], 1, None),
                grad=False, dev=dev
            )
            k_ij = lambda_n * torch.relu(k_ij)
        else:
            k_ij = lambda_n * torch.ones_like(
                to_tensor(ind), requires_grad=False, device=dev
            )

        for i in tqdm(range(map_resolution), desc='Calculating loss map', unit='row'):
            for j in range(map_resolution):
                dx, dy = xv[i, j], yv[i, j]

                rho_at_prediction = trajectory[-1, ...] / (img_shape[::-1])[None, :] * 2 - 1
                rho_at_prediction = to_tensor(rho_at_prediction, n_dim=3, dev=dev)
                rho_at_prediction[0, keypoint_to_visualize, 0] += dx / img_shape[2] * 2
                rho_at_prediction[0, keypoint_to_visualize, 1] += dy / img_shape[1] * 2

                rho_at_annotation = to_tensor(annot[target], n_dim=3, dev=dev)
                rho_at_annotation[0, keypoint_to_visualize, 0] += dx / img_shape[2] * 2
                rho_at_annotation[0, keypoint_to_visualize, 1] += dy / img_shape[1] * 2

                zephir.rho.zero_()
                zephir.rho.add_(torch.cat([rho_at_prediction, rho_at_annotation], dim=0))
                zephir.theta.zero_()

                pred = zephir(input_tensor)

                loss_r = corr_loss(pred, target_tensor)
                loss_n = reg_n(k_ij, zephir.rho, None, ind, d_ref, None, lambda_n_mode)
                loss_d = reg_d(lambda_d, input_tensor_d, zephir.rho, zephir.n_chunks)

                losses_at_trajectory[i, j, :] = np.array([
                    loss_r[0].item(), loss_n[0].item(), loss_d[0].item(),
                    loss_r[0].item() + loss_n[0].item() + loss_d[0].item()
                ])
                losses_at_annotation[i, j, :] = np.array([
                    loss_r[1].item(), loss_n[1].item(), loss_d[1].item(),
                    loss_r[1].item() + loss_n[1].item() + loss_d[1].item()
                ])

    # plot around predicted position
    xyz_at_prediction = trajectory[-1, keypoint_to_visualize, :].astype(int)
    rho_at_annotation = to_tensor(annot[target], dev=dev)
    xyz_at_annotation = np.array(
        get_pixel(rho_at_annotation[keypoint_to_visualize, :],
                  img_shape)
    )

    print(f'{keypoint_to_visualize}'
          f'\tinit: {list(trajectory[0, keypoint_to_visualize, :])}'
          f'\tfinal: {list(trajectory[-1, keypoint_to_visualize, :])}'
          f'\ttarget: {list(xyz_at_annotation)}')

    xy_target_at_prediction = map_resolution/map_size * (
        xyz_at_annotation[:2] - xyz_at_prediction[:2]
    ) + map_resolution/2
    xy_trajectory_at_prediction = map_resolution/map_size * (
        trajectory[:, keypoint_to_visualize, :2] - xyz_at_prediction[:2]
    ) + map_resolution/2

    volume_crop_at_trajectory = np.max(
        to_numpy(
            input_tensor[
             0, 0, 0,
             max(xyz_at_prediction[2] - 3, 0):xyz_at_prediction[2] + 3,
             max(xyz_at_prediction[1] - map_resolution//2, 0):xyz_at_prediction[1] + map_resolution//2,
             max(xyz_at_prediction[0] - map_resolution//2, 0):xyz_at_prediction[0] + map_resolution//2
            ]
        ), axis=0
    )

    fig, axes = plt.subplots(2, 5, constrained_layout=True, figsize=(25, 10))

    for i in range(5):
        axes[0, i].scatter(xy_trajectory_at_prediction[0, 0], xy_trajectory_at_prediction[0, 1], s=100.0, c='r')
        axes[0, i].scatter(xy_trajectory_at_prediction[-1, 0], xy_trajectory_at_prediction[-1, 1], s=100.0, c='g')
        axes[0, i].plot(xy_trajectory_at_prediction[:, 0], xy_trajectory_at_prediction[:, 1], linewidth=10.0, c='darkorange', alpha=0.3)
        axes[0, i].scatter(xy_target_at_prediction[0], xy_target_at_prediction[1], s=200.0, c='violet')
        if i == 0:
            axes[0, i].imshow(volume_crop_at_trajectory)
        else:
            axes[0, i].imshow(losses_at_trajectory[..., i-1])

    # plot around correct position
    xy_target_at_annotation = map_resolution/map_size * (
        xyz_at_annotation[:2] - xyz_at_annotation[:2]
    ) + map_resolution/2
    xy_trajectory_at_annotation = map_resolution/map_size * (
        trajectory[:, keypoint_to_visualize, :2] - xyz_at_annotation[:2]
    ) + map_resolution/2

    volume_crop_at_annotation = np.max(
        to_numpy(
            input_tensor[
             0, 0, 0,
             max(xyz_at_annotation[2] - 3, 0):xyz_at_annotation[2] + 3,
             max(xyz_at_annotation[1] - map_resolution//2, 0):xyz_at_annotation[1] + map_resolution//2,
             max(xyz_at_annotation[0] - map_resolution//2, 0):xyz_at_annotation[0] + map_resolution//2
            ]
        ), axis=0
    )

    for i in range(5):
        axes[1, i].scatter(xy_trajectory_at_annotation[0, 0], xy_trajectory_at_annotation[0, 1], s=100.0, c='r')
        axes[1, i].scatter(xy_trajectory_at_annotation[-1, 0], xy_trajectory_at_annotation[-1, 1], s=100.0, c='g')
        axes[1, i].plot(xy_trajectory_at_annotation[:, 0], xy_trajectory_at_annotation[:, 1], linewidth=10.0, c='darkorange', alpha=0.3)
        axes[1, i].scatter(xy_target_at_annotation[0], xy_target_at_annotation[1], s=200.0, c='violet')
        if i == 0:
            axes[1, i].imshow(volume_crop_at_annotation)
        else:
            axes[1, i].imshow(losses_at_annotation[..., i-1])

    plt.show()

    return losses_at_trajectory, losses_at_annotation



"""
ZephOD feature detection.

Usage:
    zephod -h | --help
    zephod -v | --version
    zephod --dataset=<dataset> --model=<model> [options]

Options:
    -h --help                           show this message and exit.
    -v --version                        show version information and exit.
    --dataset=<dataset>                 path to data directory to analyze.
    --model=<model>                     path to checkpoint for model.
    --channel=<channel>                 data channel to use as input.
    --cuda=<cuda>                       check if a CUDA-compatible GPU is available for  use. [default: True]
"""

from docopt import docopt

from ..__version__ import __version__
from ..utils.utils import *
from .model import ZephOD


def run_zephod(
        dataset=Path('.'),
        dev=torch.device('cpu'),
        channel=None,
        model_kwargs=None,
        state_dict=None):

    model = ZephOD(**model_kwargs)
    model.load_state_dict(state_dict)
    model.to(dev)
    model.eval()

    vol = get_slice(dataset, 0)
    data_file = h5py.File(dataset / 'centers.h5', 'w')
    data = data_file.create_dataset(
        'data', (0, *vol.shape[1:]),
        chunks=(1, *vol.shape[1:]),
        dtype=np.uint8,
        compression='gzip',
        compression_opts=5,
        maxshape=(None, *vol.shape[1:])
    )
    times = data_file.create_dataset(
        'times', (0,),
        chunks=(1,),
        dtype=np.float64,
        maxshape=(None,),
    )
    rec = cv2.VideoWriter(
        str(dataset / 'centers.mp4'),
        cv2.VideoWriter_fourcc(*'mp4v'), 30,
        (vol.shape[3], vol.shape[2]),
        False
    )

    t_list = get_times(dataset)
    tpbar = tqdm(t_list, desc='Analyzing movie', unit='frames')
    for t in tpbar:
        vol = get_slice(dataset, t)
        if channel is not None:
            vol = vol[channel, np.newaxis, ...]
        pred = model(vol)

        data.resize((t + 1, *vol.shape[1:]))
        data[t, ...] = to_numpy(pred[0, 0])
        times.resize((t + 1,))
        times[t] = t
        rec.write(
            np.max(
                np.transpose(to_numpy(pred[0, 0] * 255).astype(np.uint8)),
                axis=0
            )
        )
    data_file.close()
    rec.release()

    print('\n\n*** DONE!')
    return


def main():
    args = docopt(__doc__, version=f'ZephOD {__version__}')
    print(args, '\n')

    if torch.cuda.is_available() and args['--cuda'] in ['True', 'Y', 'y']:
        # Moving to GPU
        print('\n*** GPU available!')
        dev = torch.device('cuda:0')
    else:
        dev = torch.device('cpu')

    if args['--model'] is None:
        checkpoint_path = Path(__file__).parent / 'model.pt'
    else:
        checkpoint_path = Path(__file__).parent / 'models' / Path(args['--model'] + '.pt')
    if Path.is_file(checkpoint_path):
        print('\nPrevious checkpoint available.')
        try:
            checkpoint = torch.load(checkpoint_path)
        except RuntimeError:
            print('*** CUDA NOT AVAILABLE! Mapping CUDA tensors to CPU...')
            checkpoint = torch.load(checkpoint_path, map_location=torch.device('cpu'))
    else:
        print('\n******* ERROR: model not found! Exiting...')
        exit()

    run_zephod(
        dataset=Path(args['--dataset']),
        dev=dev,
        channel=int(args['--channel']) if args['--channel'] else None,
        model_kwargs=checkpoint['model_kwargs'],
        state_dict=checkpoint['state_dict'],
    )


if __name__ == '__main__':
    main()



from logging import Logger
from pathlib import Path
from typing import List

from ..data import (AnnotationTable, WorldlineTable,
                            Worldline, get_nearby_max)

from ._utilities import default_args


def create_track(
    dataset: Path,
    annotations: AnnotationTable,
    worldlines: WorldlineTable,
    window_state,
    arg: str,
    logger: Logger
) -> List[dict]:
    """
    Create a new track and select it as the current track.

    """

    w = worldlines.insert(Worldline())

    return [
        {
            "type": "worldlines/get_worldlines"
        },
        {
            "type": "annotation_window/set_selected_worldline_local",
            "payload": str(w.id)
        },
    ]



import os
from multiprocessing import Pool
from skimage.transform import resize

from ..utils.io import *


def dist_corrcoef(image_1, image_2):
    """Return a distance between two images corresponding to 1 minus the
    correlation coefficient between them. This can go from 0 to 2."""

    dist = 0
    for x1, x2 in zip(image_1, image_2):
        dist += (1 - np.corrcoef(x1.ravel(), x2.ravel())[0, 1])/len(image_1)
    return dist


def get_thumbnail(dataset, channel, t, scale):
    """Return low-resolution thumbnail of data volume."""

    v = get_slice(dataset, t)
    if channel is not None:
        v = v[channel]
    elif len(v.shape) == 4:
        v = np.max(v, axis=0)
    tmg = []
    new_shape = np.array([max(1, l//s) for l, s in zip(v.shape, scale)])
    for d in range(len(v.shape)):
        mip = np.max(v, axis=d)
        tmg.append(resize(mip, np.delete(new_shape, d)))
    return tmg


def get_all_pdists(dataset, shape_t, channel,
                   dist_fn=dist_corrcoef,
                   load=True, save=True,
                   scale=(4, 16, 16),
                   pbar=False
                   ) -> np.ndarray:
    """Return all pairwise distances between the first shape_t frames in a dataset."""

    f = dataset / 'null.npy'
    if load or save:
        if channel is not None:
            f = dataset / f'pdcc_c{channel}.npy'
        else:
            f = dataset / f'pdcc.npy'
    if f.is_file() and load:
        pdcc = np.load(str(f), allow_pickle=True)
        if pdcc.shape == (shape_t, shape_t):
            return pdcc

    print('Compiling thumbnails...')
    pool = Pool(max(os.cpu_count()-1, 1))
    thumbnails = [pool.apply_async(
        get_thumbnail, [dataset, channel, t, scale]
    ) for t in range(shape_t)]

    d = np.zeros((shape_t, shape_t))
    for i in (tqdm(range(shape_t), desc='Calculating distances', unit='frames')
              if pbar else range(shape_t)):

        for j in range(i+1, shape_t):
            dist = dist_fn(thumbnails[i].get(), thumbnails[j].get())
            if np.isnan(dist):
                d[i, j] = 2.0
            else:
                d[i, j] = dist

    pool.close()
    pool.join()

    d_full = d + np.transpose(d)
    if save:
        np.save(str(f), d_full, allow_pickle=True)

    return d_full


def get_partial_pdists(dataset, shape_t, p_list, channel,
                       dist_fn=dist_corrcoef,
                       load=True,
                       scale=(4, 16, 16),
                       pbar=False
                       ) -> np.ndarray:
    """Return pairwise distances between shape_t frames and their parents in a dataset."""

    f = dataset / 'null.npy'
    if load:
        if channel is not None:
            f = dataset / f'pdcc_c{channel}.npy'
        else:
            f = dataset / f'pdcc.npy'
    d_full = None
    if f.is_file() and load:
        d_full = np.load(str(f), allow_pickle=True)

    print('Compiling thumbnails...')
    pool = Pool(max(os.cpu_count()-1, 1))
    thumbnails = [pool.apply_async(
        get_thumbnail, [dataset, channel, t, scale]
    ) for t in range(shape_t)]

    d_partial = np.zeros(shape_t)
    for i in (tqdm(range(shape_t), desc='Calculating distances', unit='frames')
              if pbar else range(shape_t)):

        if p_list[i] < 0:
            continue

        if d_full is not None:
            d_partial[i] = d_full[i, int(p_list[i])]
        else:
            dist = dist_fn(thumbnails[i].get(), thumbnails[int(p_list[i])].get())
            if np.isnan(dist):
                d_partial[i] = 2.0
            else:
                d_partial[i] = dist

    pool.close()
    pool.join()

    return d_partial



from scipy.sparse.csgraph import shortest_path

from .build_pdists import get_all_pdists, get_partial_pdists
from ..utils.utils import *


def build_tree(
    container,
    sort_mode,
    t_ignore,
    t_track,
    verbose=False):
    """Build frame tree.

    Build frame tree with branches according to given sort_mode:
    'depth' builds branches according to minimize depth in a shortest-path search
    between reference and child.
    'similarity' builds branches to minimize distance between parent and child.
    'linear' builds branches chronologically, extending forwards and backwards
    from each reference frame.

    :param container: variable container, needs to contain: dataset, channel,
    shape_t, t_annot
    :param sort_mode: method for building frame branches
    :param t_ignore: frames to ignore for analysis
    :param t_track: frames to track for analysis; supercedes t_ignore
    :param verbose: plot relevant scores for frame tree
    :return: container (updated entries for: t_list, p_list, r_list, s_list)
    """

    # pull variables from container
    dataset = container.get('dataset')
    channel = container.get('channel')
    lr_coef = container.get('lr_coef')
    shape_t = container.get('shape_t')
    t_annot = container.get('t_annot')

    if t_track is not None:
        t_ignore = np.setdiff1d(
            np.arange(shape_t),
            np.unique(list(t_annot.copy()) + list(t_track))
        )

    d_full = None
    if (sort_mode == 'depth' or sort_mode == 'similarity') and lr_coef > 0:
        print('\nBuilding frame correlation graph...')
        d_full = get_all_pdists(dataset, shape_t, channel, pbar=True)
        if verbose:
            fig = plt.figure(figsize=(10, 10))
            plt.imshow(d_full)
            plt.show()
            fig.savefig(str(dataset / 'pw_distance.png'))
    
    if sort_mode == 'depth':
        print('\nSorting frames by depth along shortest path...')
        dist_sp, parents_sp = shortest_path(
            d_full,
            directed=False,
            return_predecessors=True,
            indices=t_annot
        )
        for i, t in enumerate(t_annot):
            parents_sp[i, t] = -1
        r_list = np.argmin(d_full[t_annot, :], axis=0)
        p_list = parents_sp[r_list, range(shape_t)]
        depths_sp = list(map(lambda x: get_depth(p_list, x), range(shape_t)))
        t_list = np.argsort(depths_sp)
    
    elif sort_mode == 'similarity':
        print('\nSorting frames by similarity to parent...')
        t_list = list(t_annot.copy())
        if t_ignore is not None:
            t_list = list(np.unique(t_list + list(t_ignore)))
        p_list = np.zeros((shape_t,))
        p_list[t_list] = -1

        d_temp = d_full.copy()
        np.fill_diagonal(d_temp, 2.1)  # dist_corrcoeff ranges from 0. to 2.0
        d_temp[:, t_list] = 2.1
        if t_ignore is not None:
            d_temp[list(t_ignore), :] = 2.1

        while len(t_list) < shape_t:
            parent, child = divmod(int(np.argmin(d_temp[t_list, :])), shape_t)
            t_list.append(child)
            p_list[child] = t_list[parent]
            d_temp[:, child] = 2.1
        t_list = np.array(t_list, dtype=int)
        p_list = np.array(p_list, dtype=int)
        r_list = np.argmin(d_full[t_annot, :], axis=0)

    else:
        print('\nSorting frames linearly off of reference frames...')
        temp = np.arange(shape_t)
        t_tree = [t_annot]
        p_tree = [[-1 for _ in t_annot]]
        r_tree = [[-1 for _ in t_annot]]
        for i in range(len(t_annot)):
            if t_annot[i] == 0:
                b = np.array([])
            elif i == 0:
                b = temp[:t_annot[i]]
            elif t_annot[i] == t_annot[i - 1] + 1:
                b = np.array([])
            else:
                b = temp[(t_annot[i] + t_annot[i - 1]) // 2:t_annot[i]]
            t_tree.append(b[::-1])
            p_tree.append(b[::-1] + 1)
            r_tree.append([i] * len(b))

            if t_annot[i] == shape_t - 1:
                f = np.array([])
            elif i == len(t_annot) - 1:
                f = temp[t_annot[i] + 1:]
            elif t_annot[i] == t_annot[i + 1] - 1:
                f = np.array([])
            else:
                f = temp[t_annot[i] + 1:(t_annot[i] + t_annot[i + 1]) // 2]
            t_tree.append(f)
            p_tree.append(f - 1)
            r_tree.append([i] * len(f))

        t_list = np.concatenate(t_tree).astype(int)
        p_list = np.concatenate(p_tree)[np.argsort(t_list)].astype(int)
        r_list = np.concatenate(r_tree)[np.argsort(t_list)].astype(int)
        if t_ignore is not None:
            for t in list(t_ignore):
                if p_list[t] > t > 0:
                    p = p_list[t-1] + 1
                    while p in list(t_ignore):
                        p += 1
                    p_list[t-1] = p
                elif p_list[t] < t < shape_t - 1:
                    p = p_list[t+1] - 1
                    while p in list(t_ignore):
                        p -= 1
                    p_list[t+1] = p
                p_list[t] = -1

    t_list = np.setdiff1d(t_list, t_annot, assume_unique=True)
    if t_ignore is not None:
        t_list = np.setdiff1d(t_list, np.array(t_ignore), assume_unique=True)

    if d_full is not None:
        s_list = get_undiscounted_scores_for_tree(d_full, p_list, shape_t)
    elif lr_coef > 0:
        s_list = get_partial_pdists(dataset, shape_t, p_list, channel, pbar=True)
    else:
        s_list = np.zeros(shape_t)
    print(f'\nFrames sorted with max/mean distance'
          f'\t{np.max(s_list):.4f} / {np.mean(s_list):.4f}')

    if verbose:
        scores_r = get_undiscounted_scores_for_tree(
            d_full, list(t_annot[r] for r in r_list), shape_t
        )
        indicators = []
        for t in t_annot:
            indicators.append([t, 0, 0.5])
        plot_with_indicator_v(
            [scores_r],
            indicators,
            title=f'Scores for tree (to root) (m={np.mean(scores_r):.4f})'
        )
    
        scores_sp = get_undiscounted_scores_for_tree(d_full, p_list, shape_t)
        indicators = []
        for t in t_annot:
            indicators.append([t, 0, 0.5])
        plot_with_indicator_v(
            [scores_sp],
            indicators,
            title=f'Scores for tree (to parent) (m={np.mean(scores_sp):.4f})'
        )

    # push variables to container
    container.update({
        't_list': t_list,
        'p_list': p_list,
        'r_list': r_list,
        's_list': s_list,
    })

    return container



from ..utils.utils import *


def save_movie(
    container,
    results,):
    """Visualize results in an annotated movie.

    Visualize results as coloured annotations in a movie. Frames are generated
    using the same data loading parameters used for tracking.

    :param container: variable container, needs to contain: dataset, channel,
    exclude_self, exclusive_prov, gamma, include_all, img_shape, shape_t, worldline_id
    :param results: tracking results
    """

    # pull variables from container
    dataset = container.get('dataset')
    channel = container.get('channel')
    exclude_self = container.get('exclude_self')
    exclusive_prov = container.get('exclusive_prov')
    gamma = container.get('gamma')
    include_all = container.get('include_all')
    img_shape = container.get('img_shape')
    shape_t = container.get('shape_t')
    t_annot = container.get('t_annot')
    t_list = container.get('t_list')
    worldline_id = container.get('worldline_id')

    annotation = get_annotation_df(dataset)

    # visualizing result and writing to video
    print('\nSaving prediction to video...')
    overview_rec = cv2.VideoWriter(
            str(dataset / 'annotated.mp4'),
            cv2.VideoWriter_fourcc(*'mp4v'), 10,
            (img_shape[2] + 2 * img_shape[0],
             img_shape[1] + 2 * img_shape[0]),
            True
    )

    if include_all:
        t_list = np.arange(shape_t)
    else:
        t_list = np.unique(list(t_annot) + list(t_list))

    for t in tqdm(t_list, desc='Saving to video', unit='frames'):
        data = get_data(dataset, t, g=gamma, c=channel)
        xyz_result = results[t, ...]
        if include_all and t in np.unique(annotation['t_idx']):
            u, annot, prov = get_annotation(annotation, t, exclusive_prov, exclude_self)
            w_idx = np.where(np.isin(worldline_id, u))[0]
            if len(w_idx) > 0:
                u_idx = np.where(np.isin(u, worldline_id))[0]
                xyz_result[w_idx, ...] = annot[u_idx, ...]
            if len(u) > len(w_idx):
                u_idx = np.where(np.isin(u, worldline_id, invert=True))[0]
                xyz_result = np.append(xyz_result, annot[u_idx, ...], axis=0)
        overview = save_as_bgr(data, c=channel, annotation=xyz_result)
        cv2.putText(overview, f'Frame #{t}', (20, 20),
                    cv2.FONT_HERSHEY_PLAIN, 1, (255, 255, 255), thickness=1)
        overview_rec.write(overview[:, :, ::-1])
    overview_rec.release()

    return



__version__ = '1.0.3'
# from pathlib import Path
# import subprocess
#
# __version__ = subprocess.check_output(
#     ["git", "describe", "--tags"],
#     cwd=Path(__file__).parent.parent.absolute()
# ).strip().decode('UTF-8')



"""
recommend_frames.py: search and determine optimal reference frames to annotate for ZephIR.

Determine median frames to recommend as reference frames via k-medoids clustering
based on thumbnail distances (see build_pdists). Clustering is done iteratively,
such that one cluster is determined at a time. n_iter > 0 will re-iterate over
the existing list of median frames to fine-tune recommendations.

Usage:
    recommend_frames.py -h | --help
    recommend_frames.py -v | --version
    recommend_frames.py --dataset=<dataset> [options]

Options:
    -h --help                           	show this message and exit.
    -v --version                        	show version information and exit.
    --dataset=<dataset>  					path to data directory to analyze.
    --n_frames=<n_frames>  					number of reference frames to search for. [default: 5]
    --n_iter=<n_iter>  						number of iterations for optimizing results; -1 to uncap. [default: 0]
    --t_list=<t_list>  						frames to analyze.
    --channel=<channel>  					data channel to use for calculating correlation coefficients.
    --save_to_metadata=<save_to_metadata>  	save t_ref to metadata.json. [default: True]
    --verbose=<verbose>  					return score plots during search. [default: False]
"""

from collections import OrderedDict
from docopt import docopt

from ..__version__ import __version__
from ..methods.build_pdists import get_all_pdists
from ..utils.utils import *


def recommend_frames(
    dataset, n_frames, n_iter, t_list, channel,
    save_to_metadata, verbose
):

    metadata = get_metadata(dataset)
    shape_t = metadata['shape_t']
    if t_list is None:
        t_list = list(range(shape_t))

    print('Building frame correlation graph...')
    d_full = get_all_pdists(dataset, shape_t, channel, pbar=True)
    d_slice = (d_full[t_list, :])[:, t_list]

    scores = np.mean(d_slice, axis=-1)
    opt_score, med_idx = np.min(scores), np.argmin(scores)

    i_ref = [med_idx]
    t_ref = [t_list[med_idx]]
    s_ref = [opt_score]
    scores = d_slice[med_idx, :]
    pbar = tqdm(range(n_frames - 1), desc='Optimizing reference frames', unit='n_frames')
    for i in pbar:
        d_adj = np.append(
            d_slice.copy()[:, :, None],
            np.tile(scores[None, :, None], (d_slice.shape[0], 1, 1)),
            axis=-1
        )
        d_opt = np.min(d_adj, axis=-1)
        iscores = np.mean(d_opt, axis=-1)
        opt_score, new_midx = np.min(iscores), np.argmin(iscores)

        i_ref.append(new_midx)
        t_ref.append(t_list[new_midx])
        s_ref.append(opt_score)
        scores = d_opt[new_midx, :]
    print(f'\nFirst pass reference frames: {t_ref}'
          f'\nCurrent optimized score: {opt_score:.4f}')

    if verbose:
        plot_with_indicator_v(
            [scores],
            [[t, np.min(scores), 0.5] for t in t_ref],
            x_list=[t_list],
            title=f'First pass mean score: {opt_score:.4f}'
        )

    print(f'\nIterating over found reference frames...')
    n_i = 0
    while True:
        if 0 <= n_iter <= n_i:
            break
        kscore = opt_score
        for i in range(n_frames):
            i_ref_temp = i_ref.copy()
            i_ref_temp.pop(i)
            d_adj = d_slice.copy()[:, :, None]
            for t in i_ref_temp:
                d_adj = np.append(
                    d_adj,
                    np.tile(d_slice.copy()[t, None, :, None],
                            (d_adj.shape[0], 1, 1)),
                    axis=-1
                )

            iscores = np.mean(np.min(d_adj, axis=-1), axis=-1)
            jscore, new_midx = np.min(iscores), np.argmin(iscores)

            if jscore < kscore:
                i_ref[i] = new_midx
                t_ref[i] = t_list[new_midx]
                kscore = jscore

        if kscore < opt_score:
            opt_score = kscore
            n_i += 1
            if verbose:
                print(f'\nIter#{n_i}\tCurrent reference frames: {t_ref}'
                      f'\t\tCurrent optimal score: {opt_score:.4f}')
        else:
            break
    print(f'\nFinal reference frames: {t_ref}\n'
          f'Final optimized score: {opt_score:.4f}')

    if verbose:
        plot_with_indicator_v(
            [scores],
            [[t, np.min(scores), 0.5] for t in t_ref],
            x_list=[t_list],
            title=f'Final mean score: {opt_score:.4f}'
        )

        plt.figure()
        # plt.title('opt_score vs N_ref')
        plt.xlabel('Number of reference frames')
        plt.ylabel('Mean distance to reference')
        plt.ylim(0, np.max(s_ref)+0.05)
        plt.plot(np.arange(1, len(s_ref) + 1), s_ref)
        plt.show()

    if save_to_metadata:
        update_metadata(dataset, {f't_ref_fn{len(t_list)}': [int(i) for i in t_ref]})


def main():
    args = docopt(__doc__, version=f'ZephIR recommend_frames {__version__}')
    print(args, '\n')

    recommend_frames(
        dataset=Path(args['--dataset']),
        n_frames=int(args['--n_frames']),
        n_iter=int(args['--n_iter']),
        t_list=eval(args['--t_list']) if args['--t_list'] else None,
        channel=int(args['--channel']) if args['--channel'] else None,
        save_to_metadata=args['--save_to_metadata'] in ['True', 'Y', 'y'],
        verbose=args['--verbose'] in ['True', 'Y', 'y'],
    )


if __name__ == '__main__':
    main()



from ..utils.utils import *

from ..models.zephir import ZephIR
from ..zephod.model import ZephOD


def build_models(
    container,
    dimmer_ratio,
    grid_shape,
    fovea_sigma,
    n_chunks,):
    """Compile PyTorch models for tracking.

    Models are built according to user arguments. Zephod is only built when a
    compatible model is available.

    :param container: variable container, needs to contain: dataset, allow_rotation,
    dev, img_shape, n_frame, shape_n
    :param dimmer_ratio: Zephir parameter, dims descriptor edges relative to center
    :param grid_shape: Zephir parameter, descriptor size
    :param fovea_sigma: Zephir parameter, size of foveated region
    :param n_chunks: Zephir parameter, number of chunks to divide forward pass into
    :return: container (updated entry for: grid_shape), zephir, zephod
    """
    # pull variables from container
    dataset = container.get('dataset')
    allow_rotation = container.get('allow_rotation')
    dev = container.get('dev')
    img_shape = container.get('img_shape')
    n_frame = container.get('n_frame')
    shape_n = container.get('shape_n')
    
    z_stacks = min(img_shape[0], grid_shape[0])
    grid_shape = (z_stacks - (z_stacks+1) % 2, grid_shape[1], grid_shape[2])
    grid_spacing = tuple(np.array(grid_shape) / np.array(img_shape))
    
    # models and loss function compiled
    print(f'\nCompiling models and loss function...')
    model_kwargs = {
        'allow_rotation': allow_rotation,
        'dimmer_ratio': dimmer_ratio,
        'fovea_sigma': fovea_sigma,
        'grid_shape': grid_shape,
        'grid_spacing': grid_spacing,
        'n_chunks': n_chunks,
        'n_frame': n_frame,
        'shape_n': shape_n,
        'ftr_ratio': 0.6,
        'ret_stride': 2,
    }
    zephir = ZephIR(**model_kwargs)

    # compiling Zephod models from existing kwargs and state_dict
    zephod_path = dataset / 'zephod.pt'
    if not zephod_path.is_file():
        zephod_path = Path(__file__).parent.parent / 'zephod' / 'models.pt'
    if zephod_path.is_file():
        try:
            zephod_checkpoint = torch.load(str(zephod_path))
        except RuntimeError:
            print('*** CUDA NOT AVAILABLE! Mapping Zephod parameters to CPU...')
            zephod_checkpoint = torch.load(str(zephod_path), map_location=torch.device('cpu'))
        zephod = ZephOD(**zephod_checkpoint['model_kwargs'])
        zephod.load_state_dict(zephod_checkpoint['state_dict'])
        zephod.to(dev)
    else:
        print('******* ERROR: Zephod checkpoint file not found!')
        zephod = None

    # push variables to container
    container.update({
        'grid_shape': grid_shape,
    })

    # push models to checkpoint
    update_checkpoint(dataset, {
        'zephir_kwargs': model_kwargs,
        'zephir': zephir,
        'zephod': zephod
    })

    return container, zephir, zephod



from setuptools import setup, find_packages
# from .zephir.__version__ import __version__
__version__ = '1.0.3'

requirements = [
    'dataclasses>=0.6'
    'docopt>=0.6.2',
    'Flask>=2.1.2',
    'gevent>=21.12.0',
    'h5py>=3.6.0',
    'matplotlib>=3.5.2',
    'numpy>=1.22.4',
    'opencv-python>=4.5.5.64'
    'pandas>=1.4.2',
    'scikit-image>=0.19.2',
    'scikit-learn>=1.0.2',
    'scipy>=1.7.3',
    'setuptools>=61.2.0',
    'torch>=1.10.0',
    'tqdm>=4.64.0',
]

setup(
    name='zephir',
    version=__version__,
    license='MIT',
    description='Multiple object tracking algorithm via image registration',
    author='James Yu, Vivek Venkatachalam',
    author_email='v.venkatachalam@northeastern.edu',
    maintainer='James Yu',
    maintainer_email='jd.ryu@icloud.com',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/venkatachalamlab/ZephIR',
    entry_points={'console_scripts': [
        'zephir=zephir.main:main',
        'annotator=zephir.annotator.main:main',
        'auto_annotate=zephir.methods.auto_annotate:main',
        'extract_traces=zephir.methods.extract_traces:main',
        'overwrite_checkpoint=zephir.methods.overwrite_checkpoint:main',
        'recommend_frames=zephir.methods.recommend_frames:main',
        'train_zephod=zephir.zephod.train:main',
        'zephod=zephir.zephod.main:main',
    ]},
    include_package_data=True,
    zip_safe=False,
    # keywords=[
    #     'image registration',
    #     'neuron tracking',
    #     'object tracking',
    #     'multiple object tracking',
    # ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
    ],
    install_requires=requirements,
    packages=find_packages()
)



"""
To ease the pain of ensuring compatibility with new data structures or datasets,
this file collects key IO functions for data, metadata, and annotations
that may be edited by a user to fit their particular use case.
"""

import h5py
import json
import numpy as np
import os
import pandas as pd
from pathlib import Path


# default getters
def get_slice(dataset: Path, t: int) -> np.ndarray:
    """Return a slice at specified index t.
    This should return a 4-D numpy array containing multi-channel volumetric data
    with the dimensions ordered as (C, Z, Y, X).
    """
    h5_filename = dataset / "data.h5"
    f = h5py.File(h5_filename, 'r')
    return f["data"][t]


def get_annotation_df(dataset: Path) -> pd.DataFrame:
    """Load and return annotations as an ordered pandas dataframe.
    This should contain the following:
    - t_idx: time index of each annotation
    - x: x-coordinate as a float between (0, 1)
    - y: y-coordinate as a float between (0, 1)
    - z: z-coordinate as a float between (0, 1)
    - worldline_id: track or worldline ID as an integer
    - provenance: scorer or creator of the annotation as a byte string
    """
    with h5py.File(dataset / 'annotations.h5', 'r') as f:
        data = pd.DataFrame()
        for k in f:
            data[k] = f[k]
    return data


def get_metadata(dataset: Path) -> dict:
    """Load and return metadata for the dataset as a Python dictionary.
    This should contain at least the following:
    - shape_t
    - shape_c
    - shape_z
    - shape_y
    - shape_x
    """
    json_filename = dataset / "metadata.json"
    with open(json_filename) as json_file:
        metadata = json.load(json_file)
    return metadata



from typing import Tuple

import cv2
import numpy as np

from .annotations_io import AnnotationTable, WorldlineTable
from .transform import mip_threeview


def hex_to_rgb(value: bytes) -> Tuple[int, ...]:
    value = value.lstrip(b'#')
    lv = len(value)
    return tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))


def draw_on_threeview(A: AnnotationTable,
                      W: WorldlineTable,
                      V: np.ndarray,
                      dot_size=2,
                      scale=(4, 1, 1)):

    threeview = mip_threeview(V, scale=scale)

    size_X = round(scale[2] * V.shape[2])
    size_Y = round(scale[1] * V.shape[1])
    size_Z = round(scale[0] * V.shape[0])

    for a in A:

        x = (a.x * size_X).astype(int)
        y = (a.y * size_Y).astype(int)
        z = (a.z * size_Z).astype(int)
        color = hex_to_rgb(W.get(a.worldline_id).color)

        # Draw on z-projection
        cv2.circle(threeview, (x, y), dot_size, color, -1)

        # Draw on x-projection
        cv2.circle(threeview, (size_X + z, y), dot_size, color, -1)

        # Draw on y-projection
        cv2.circle(threeview, (x, size_Y + z), dot_size, color, -1)

    return threeview


"""
Training ZephOD feature detection network

Usage:
    train_zephod -h | --help
    train_zephod -v | --version
    train_zephod --index=<index> [options]

Options:
    -h --help                           show this message and exit.
    -v --version                        show version information and exit.
    --index=<index>                     path to json index of data directories to use as training data.
    --batch_size=<batch_size>           batch size in volumes. [default: 1]
    --channel=<channel>                 data channel to use as input.
    --cuda=<cuda>                       check if a CUDA-compatible GPU is available for  use. [default: True]
    --loss_func=<loss_func>             loss function to use. [default: BCE]
    --lr_init=<lr_init>                 learning rate for optimizer. [default: 1.0]
    --model=<model>                     name of model being trained.
    --n_epoch=<n_epoch>                 number of epochs to train for. [default: 64]
    --n_trn=<n_trn>                     number of training samples per epoch. [default: 50]
    --n_val=<n_val>                     number of validation samples per epoch. [default: 5]
    --override=<override>               override existing checkpoint. [default: False]
"""

from docopt import docopt
import random
import shutil
import torch.nn as nn
import torch.optim as optim

from ..__version__ import __version__
from ..utils.getters import *
from zephir.models.losses import corr_loss
from ..utils.utils import *
from .augment import generate_synthetic_data, identify
from .model import ZephOD


def train_model(
        index=Path('.'),
        dev=torch.device('cpu'),
        channel=None,
        n_epoch=10,
        loss_func='BCE',
        lr_init=0.1,
        batch_size=1,
        n_trn=100,
        n_val=10,
        model_kwargs=None,
        state_dict=None):

    with open(index) as f:
        _index = json.load(f)
        dataset_paths = [Path(p) for p in _index.values()]

    annots_agg = []
    t_note_agg = []
    isolates_agg = []
    for dataset in dataset_paths:
        annotation = get_annotation_df(dataset)
        t_note = np.unique(annotation['t_idx']).astype(int)
        t_note = np.sort(t_note)
        nn_list = [len(get_annotation(annotation, t)[0]) for t in t_note]
        n_neuron, t_neuron = np.max(nn_list), t_note[np.argmax(nn_list)]
        worldline_id, _, _ = get_annotation(annotation, t_neuron)
        print(f'\nUsing frame #{t_neuron} as initial reference with {n_neuron} annotations found...')

        metadata = get_metadata(dataset)
        shape_t = metadata['shape_t']
        xyz_note = []
        for t in t_note:
            u, annot, prov = get_annotation(annotation, t)
            u_idx = np.array([np.where(u == w)[0][-1] for w in worldline_id if w in u], dtype=int)
            annot = annot[u_idx, ...]
            if (t >= shape_t
                    or annot.shape[0] > n_neuron
                    or annot.shape[0] < n_neuron
                    or annot.shape[0] == 0
                ):
                t_note = np.setdiff1d(t_note, [t])
                continue
            xyz_note.append(annot)
        print(f'Annotations loaded for frames {t_note} with shape: {np.array(xyz_note).shape}')

        annots = np.array(xyz_note)
        annots_agg.append(annots)
        t_note_agg.append(t_note)

        pbar = tqdm(
            range(len(t_note)),
            desc='Finding isolated features',
            unit='frames',
        )
        for i in pbar:
            # for i in range(len(t_note)):
            data = get_data(dataset, t_note[i], c=channel)
            isolated_neurons = identify(data, annots[i], 12)
            for n in isolated_neurons:
                isolates_agg.append(n)
            pbar.set_postfix(Neurons=f'{len(isolates_agg)}')

    print(f'\n\nAnnotations loaded for a total of {len(annots_agg)} datasets!'
          f'\n*** Found a total of {len(isolates_agg)} isolated neurons')

    print('\nCompiling model...')
    vol = get_data(dataset_paths[0], t_note_agg[0][0], c=channel)
    if model_kwargs is None:
        model_kwargs = {
            'n_channels_in': 5 * vol.shape[0],
            'n_channels_out': 1,
            'init_nodes': 16,
            'kernel': (1, 3, 3),
            'padding': 1,
            'pool_kernel': (2, 2, 2)
        }
    model = ZephOD(**model_kwargs).to(dev)
    if state_dict is not None:
        print('Loading model from existing state_dict...')
        model.load_state_dict(state_dict)
    optimizer = optim.Adadelta(model.parameters(), lr=lr_init)
    scaler = torch.cuda.amp.GradScaler()
    if loss_func == 'BCE':
        loss_function = nn.BCEWithLogitsLoss()
    else:
        loss_function = corr_loss

    n_val = max(min(n_val, len(t_note_agg) - 1), 0)
    t_list = t_note_agg[:-n_val]
    t_val = t_note_agg[-n_val:]

    print('\n\n\n******* BEGIN TRAINING *******\n')
    model.train()
    pbar = tqdm(range(n_epoch), desc='Training model', unit='epochs')
    for epoch in pbar:
        # for epoch in range(n_epoch):
        tpbar = tqdm(range(n_trn), leave=False, desc='Training samples', unit='vol')
        for _ in tpbar:
            # for i in range(n_trn):
            with torch.no_grad():
                d_idx = random.randint(0, len(t_list) - 1)
                dataset = dataset_paths[d_idx]
                vol = get_data(dataset, t_note_agg[d_idx][0], c=channel)

                input_list, target_list = [], []
                for j in range(batch_size):
                    t_idx = random.randint(0, len(t_note_agg[d_idx]) - 1)
                    synth, labels = generate_synthetic_data(
                        vol, annots_agg[d_idx][t_idx], isolates_agg, True, True
                    )
                    input_list.append(to_tensor(synth, dev=dev))
                    target_list.append(to_tensor(labels, dev=dev))
                input_tensor, target_tensor = torch.stack(input_list, dim=0).to(torch.float16), torch.stack(target_list, dim=0).to(torch.float16)

            optimizer.zero_grad()
            with torch.autocast(device_type=dev):
                pred = model(input_tensor)
                loss = loss_function(pred, target_tensor)
            # loss.backward()
            # optimizer.step()
            scaler.scale(loss).backward()
            scaler.step(optimizer)
            scaler.update()

            with torch.no_grad():
                tpbar.set_postfix(Loss=f'{loss.item():.4f}')
                pbar.set_postfix(Loss=f'{loss.item():.4f}')
    
    with torch.no_grad():
        checkpoint = {
            'model_kwargs': model_kwargs,
            'state_dict': model.state_dict(),
            'opt_dict': optimizer.state_dict(),
        }

        input_val = target_val = val_pred = None
        if n_val > 0:
            model.eval()
            input_val, target_val = [], []
            for i in tqdm(range(len(t_val)), leave=False, desc='Validating samples', unit='vol'):
                # for i in range(len(t_val)):
                dataset = dataset_paths[-n_val + i]
                vol = get_data(dataset, t_note_agg[-n_val + i][0], c=channel)
                for j in range(len(t_note_agg[-n_val + i])):
                    synth, labels = generate_synthetic_data(
                        vol, annots_agg[-n_val + i][j], isolates_agg, True, True
                    )
                    input_val.append(to_tensor(synth, dev=dev))
                    target_val.append(to_tensor(labels, dev=dev))
            input_val, target_val = torch.stack(input_val, dim=0), torch.stack(target_val, dim=0)
            val_pred = model(input_val)
            val_loss = loss_function(val_pred, target_val)
            print(f'\nValidation loss: {val_loss.item():.4f}\n\n')

    return checkpoint, (input_val, target_val, val_pred), (to_numpy(model.alpha), to_numpy(model.gamma))


def main():
    args = docopt(__doc__, version=f'ZephOD {__version__}')
    print(args, '\n')

    if torch.cuda.is_available() and args['--cuda'] in ['True', 'Y', 'y']:
        # Moving to GPU
        print('\n*** GPU available!\n')
        dev = 'cuda'
    else:
        dev = 'cpu'

    if args['--model'] is None:
        name = 'model'
    else:
        name = args['--model']
    checkpoint_path = Path(__file__).parent / f'{name}.pt'

    override = args['--override'] in ['True', 'Y', 'y']
    checkpoint = {
        'model_kwargs': None,
        'state_dict': None,
        'opt_dict': None,
        'loss_list': None
    }
    if Path.is_file(checkpoint_path):
        print('\nPrevious checkpoint available.')
        if not (checkpoint_path.parent / 'backup').is_dir():
            Path.mkdir(checkpoint_path.parent / 'backup')
        now = datetime.datetime.now()
        _now = now.strftime("%m_%d_%Y_%H_%M")
        shutil.copy(
            checkpoint_path,
            checkpoint_path.parent / 'backup' / f'{name}_{_now}.pt'
        )
        if not override:
            try:
                checkpoint = torch.load(checkpoint_path)
            except RuntimeError:
                print('*** CUDA NOT AVAILABLE! Mapping CUDA tensors to CPU...')
                checkpoint = torch.load(checkpoint_path, map_location=torch.device('cpu'))

    checkpoint, validation, params = train_model(
        index=Path(args['--index']),
        dev=dev,
        channel=int(args['--channel']) if args['--channel'] else None,
        n_epoch=int(args['--n_epoch']),
        loss_func=str(args['--loss_func']),
        lr_init=float(args['--lr_init']),
        batch_size=int(args['--batch_size']),
        n_trn=int(args['--n_trn']),
        n_val=int(args['--n_val']),
        model_kwargs=checkpoint['model_kwargs'],
        state_dict=checkpoint['state_dict'],
    )

    print(f'\nSaving checkpoint to: {checkpoint_path.as_posix()}')
    torch.save(checkpoint, checkpoint_path)

    input_val, target_val, val_pred = validation
    if val_pred is not None:
        print('\nSaving prediction to a new file...')
        if not (checkpoint_path.parent / 'bin').is_dir():
            Path.mkdir(checkpoint_path.parent / 'bin')
        fig, axes = plt.subplots(1, 3, figsize=(24, 6))
        axes[0].imshow(np.max(np.max(to_numpy(input_val[0]), axis=0), axis=0), vmin=0, vmax=1)
        axes[0].set_title('Input Volume')
        axes[1].imshow(np.max(to_numpy(val_pred[0][0]), axis=0), vmin=0, vmax=1)
        axes[1].set_title('Output Labels')
        im = axes[2].imshow(np.max(to_numpy(target_val[0][0]), axis=0), vmin=0, vmax=1)
        axes[2].set_title('GT Labels')
        fig.colorbar(im, ax=axes.ravel().tolist())
        now = datetime.datetime.now()
        _now = now.strftime("%m_%d_%Y_%H_%M")
        plt.savefig(str(checkpoint_path.parent / 'bin' / f'{name}_pred_{_now}.png'))

    print('\n\n*** DONE!\n')


if __name__ == '__main__':
    main()



import sklearn.neighbors as sk

from ..utils.utils import *


def build_springs(
    container,
    load_nn,
    nn_max,
    verbose=False,):
    """Build nearest-neighbor spring network for spatial regularization.

    Build KDTree of nn_max nearest neighbors around each annotation. These connections
    are used to calculate spatial regularization loss, L_N. Covariances of
    connected pairs are compiled if multiple reference frames are available.

    :param container: variable container, needs to contain: dataset, annot,
    grid_shape, img_shape, shape_n, t_annot, verbose, z_compensator
    :param load_nn: load or save existing nn_idx.txt for spring connection indices
    :param nn_max: maximum number of neighbors to connect to for each keypoint
    :return: container (updated entries for: covar, neighbors)
    """

    # pull variables from container
    dataset = container.get('dataset')
    annotations = container.get('annot')
    grid_shape = container.get('grid_shape')
    img_shape = container.get('img_shape')
    shape_n = container.get('shape_n')
    t_annot = container.get('t_annot')
    z_compensator = container.get('z_compensator')

    print('\nBuilding neighbor tree...')
    d_scaler = np.array([1, 1, z_compensator])
    weighted_shape = np.array(img_shape)[::-1] * np.array([1, 1, z_compensator])

    covar = None
    if len(t_annot) > 1:
        # getting variance of intra-keypoint distances from annotations
        covar = np.zeros((shape_n, shape_n))
        for dim in range(3):
            covar += np.cov(
                (np.array(annotations)[:, :, dim]+1)/2 * weighted_shape[dim],
                rowvar=False) / 3

        if verbose:
            fig = plt.figure(figsize=(10, 10))
            fig.suptitle(f'mean={np.mean(covar):.3f}  '
                         f'max={np.max(covar):.3f}  '
                         f'min={np.min(covar):.3f}', fontsize=14)
            c = covar + np.min(covar)
            plt.imshow(c / np.max(c))
            plt.show()
            fig.savefig(str(dataset / 'pw_covariance.png'))

    # neighbors are filtered by position covariance
    nn_max = min(nn_max, shape_n-1)
    if load_nn and (dataset / 'nn_idx.txt').is_file():
        neighbors = []
        with open(str(dataset / 'nn_idx.txt'), 'r') as k:
            nn_idx = [line.rstrip().split(' ') for line in k.readlines()]
        nn_max = np.max([len(k) for k in nn_idx])
        ind = np.tile(np.arange(shape_n)[:, None], (1, nn_max))
        for nn in nn_idx:
            ind[int(nn[0]), :len(nn[1:])] = np.array(nn[1:]).astype(int)
        for annot in annotations:
            d_ref = (np.concatenate([annot[ind[:, k], np.newaxis, :]
                                     for k in range(nn_max)], axis=1)
                     - np.tile(annot[:, np.newaxis, :], (1, nn_max, 1)))
            neighbors.append(np.append(ind[:, :, np.newaxis], d_ref, axis=-1))

    elif nn_max > 0:
        neighbors = []
        for annot in annotations:
            weighted_xyz = (annot + 1)/2 * weighted_shape
            tree = sk.KDTree(weighted_xyz)
            dist, ind = tree.query(weighted_xyz, k=nn_max+1, return_distance=True)
            ind = np.where(
                dist[:, 1:] < 1.6/np.prod(d_scaler) * np.linalg.norm(grid_shape*d_scaler),
                ind[:, 1:],
                np.tile(np.arange(shape_n)[:, None], (1, nn_max))
            )
            d_ref = (np.concatenate([annot[ind[:, k], np.newaxis, :]
                                     for k in range(nn_max)], axis=1)
                     - np.tile(annot[:, np.newaxis, :], (1, nn_max, 1)))
            neighbors.append(np.append(ind[:, :, np.newaxis], d_ref, axis=-1))
            if load_nn:
                np.savetxt(
                    str(dataset / 'nn_idx.txt'),
                    np.append(np.arange(shape_n)[:, None], ind, axis=-1),
                    fmt='%d'
                )

    else:
        neighbors = [np.append(
            np.arange(0, shape_n)[:, None, None],
            -1 * np.ones((shape_n, 1, 3)),
            axis=-1) for _ in t_annot]

    if verbose:
        data = get_data(dataset, t_annot[0])
        overview = save_as_bgr(data, annotation=annotations[0])
        for n in range(shape_n):
            xn, yn, _ = get_pixel(annotations[0][n, ...], img_shape)
            ind = neighbors[0][n, :, 0].astype(int)
            for i in ind:
                xi, yi, _ = get_pixel(annotations[0][i, ...], img_shape)
                overview = cv2.line(overview, (xn, yn), (xi, yi),
                                    color=pick_color(n))
        fig = plt.figure(dpi=600)
        plt.imshow(overview)
        plt.show()
        fig.savefig(str(dataset / f'skeleton #{t_annot[0]}.png'))

    # push variables to container
    container.update({
        'covar': covar,
        'neighbors': neighbors,
    })

    return container
